@@ -226,7 +226,7 @@ GGG";
 
 
 ModelCodon::ModelCodon(const char *model_name, string model_params, StateFreqType freq, string freq_params,
-		PhyloTree *tree, bool count_rates) : ModelGTR(tree, count_rates)
+		PhyloTree *tree) : ModelMarkov(tree)
 {
     half_matrix = false;
     omega = kappa = kappa2 = 1.0;
@@ -262,8 +262,12 @@ ModelCodon::~ModelCodon() {
 	}
 }
 
-void ModelCodon::saveCheckpoint() {
+void ModelCodon::startCheckpoint() {
     checkpoint->startStruct("ModelCodon");
+}
+
+void ModelCodon::saveCheckpoint() {
+    startCheckpoint();
 //    CKP_ARRAY_SAVE(12, ntfreq);
     CKP_SAVE(omega);
 //    CKP_SAVE(fix_omega);
@@ -275,13 +279,14 @@ void ModelCodon::saveCheckpoint() {
 //    CKP_SAVE(fix_kappa2);
 //    int codon_freq_style = this->codon_freq_style;
 //    CKP_SAVE(codon_freq_style);
-    checkpoint->endStruct();
-    ModelGTR::saveCheckpoint();
+    endCheckpoint();
+
+    ModelMarkov::saveCheckpoint();
 }
 
 void ModelCodon::restoreCheckpoint() {
-    ModelGTR::restoreCheckpoint();
-    checkpoint->startStruct("ModelCodon");
+    ModelMarkov::restoreCheckpoint();
+    startCheckpoint();
 //    CKP_ARRAY_RESTORE(12, ntfreq);
     CKP_RESTORE(omega);
 //    CKP_RESTORE(fix_omega);
@@ -295,7 +300,7 @@ void ModelCodon::restoreCheckpoint() {
 //    int codon_freq_style;
 //    CKP_RESTORE(codon_freq_style);
 //    this->codon_freq_style = (CodonFreqStyle)codon_freq_style;
-    checkpoint->endStruct();
+    endCheckpoint();
 
     decomposeRateMatrix();
     if (phylo_tree)
@@ -345,8 +350,9 @@ StateFreqType ModelCodon::initCodon(const char *model_name, StateFreqType freq,
 		return FREQ_USER_DEFINED;
 	} else {
 		//cout << "User-specified model "<< model_name << endl;
-		readParameters(model_name);
+		//readParameters(model_name);
 			//name += " (user-defined)";
+        readCodonModelFile(model_name, reset_params);
 		return FREQ_USER_DEFINED;
 	}
 
@@ -410,7 +416,7 @@ void ModelCodon::init(const char *model_name, string model_params, StateFreqType
 		//ntfreq = new double[12];
 		phylo_tree->aln->computeCodonFreq(freq, state_freq, ntfreq);
 	}
-	ModelGTR::init(freq);
+	ModelMarkov::init(freq);
 }
 
 StateFreqType ModelCodon::initMG94(bool fix_kappa, StateFreqType freq, CodonKappaStyle kappa_style) {
@@ -491,21 +497,23 @@ void ModelCodon::computeRateAttributes() {
             int nuc1, nuc2;
             int attr = 0;
             ts = tv = 0;
-            if (phylo_tree->aln->genetic_code[i] == phylo_tree->aln->genetic_code[j])
+            int codoni = phylo_tree->aln->codon_table[i];
+            int codonj = phylo_tree->aln->codon_table[j];
+            if (phylo_tree->aln->genetic_code[codoni] == phylo_tree->aln->genetic_code[codonj])
                 attr |= CA_SYNONYMOUS;
             else
                 attr |= CA_NONSYNONYMOUS;
                 
                 
-            int nt_changes = ((i/16) != (j/16)) + (((i%16)/4) != ((j%16)/4)) + ((i%4) != (j%4));
-            int aa1 = strchr(symbols_protein, phylo_tree->aln->genetic_code[i]) - symbols_protein;
-            int aa2 = strchr(symbols_protein, phylo_tree->aln->genetic_code[j]) - symbols_protein;
-            assert(aa1 >= 0 && aa1 < 20 && aa2 >= 0 && aa2 < 20);
+            int nt_changes = ((codoni/16) != (codonj/16)) + (((codoni%16)/4) != ((codonj%16)/4)) + ((codoni%4) != (codonj%4));
+            int aa1 = strchr(symbols_protein, phylo_tree->aln->genetic_code[codoni]) - symbols_protein;
+            int aa2 = strchr(symbols_protein, phylo_tree->aln->genetic_code[codonj]) - symbols_protein;
+            ASSERT(aa1 >= 0 && aa1 < 20 && aa2 >= 0 && aa2 < 20);
             if (nt_changes < aa_cost_change[aa1*20+aa2]) {
                 aa_cost_change[aa1*20+aa2] = aa_cost_change[aa2*20+aa1] = nt_changes;
             }
             
-            if ((nuc1=i/16) != (nuc2=j/16)) {
+            if ((nuc1=codoni/16) != (nuc2=codonj/16)) {
                 if (abs(nuc1-nuc2)==2) { // transition 
                     attr |= CA_TRANSITION_1NT;
                     ts++;
@@ -514,7 +522,7 @@ void ModelCodon::computeRateAttributes() {
                     tv++;
                 }
             }
-            if ((nuc1=(i%16)/4) != (nuc2=(j%16)/4)) {
+            if ((nuc1=(codoni%16)/4) != (nuc2=(codonj%16)/4)) {
                 if (abs(nuc1-nuc2)==2) { // transition
                     attr |= CA_TRANSITION_2NT;
                     ts++;
@@ -523,7 +531,7 @@ void ModelCodon::computeRateAttributes() {
                     tv++;
                 }
             }
-            if ((nuc1=i%4) != (nuc2=j%4)) {
+            if ((nuc1=codoni%4) != (nuc2=codonj%4)) {
                 if (abs(nuc1-nuc2)==2) { // transition
                     attr |= CA_TRANSITION_3NT;
                     ts++;
@@ -579,15 +587,18 @@ void ModelCodon::combineRateNTFreq() {
         for (j = 0; j < num_states; j++)  {
             if (this_rate[j] == 0.0)
                 continue;
+
+            int codoni = phylo_tree->aln->codon_table[i];
+            int codonj = phylo_tree->aln->codon_table[j];
             int nuc1, nuc2;
                 
-            if ((nuc1=i/16) != (nuc2=j/16)) {
+            if ((nuc1=codoni/16) != (nuc2=codonj/16)) {
                 this_rate[j] *= ntfreq[nuc2];
             }
-            if ((nuc1=(i%16)/4) != (nuc2=(j%16)/4)) {
+            if ((nuc1=(codoni%16)/4) != (nuc2=(codonj%16)/4)) {
                 this_rate[j] *= ntfreq[nuc2+4];
             }
-            if ((nuc1=i%4) != (nuc2=j%4)) {
+            if ((nuc1=codoni%4) != (nuc2=codonj%4)) {
                 this_rate[j] *= ntfreq[nuc2+8];
             }
         }
@@ -627,8 +638,8 @@ void ModelCodon::readCodonModel(istream &in, bool reset_params) {
 		int nt3 = phylo_tree->aln->convertState(codons[i][2], SEQ_DNA);
 		if (nt1 > 3 || nt2 > 3 || nt3 > 3)
 			outError("Wrong codon triplet ", codons[i]);
-		state_map[i] = nt1*16+nt2*4+nt3;
-		if (phylo_tree->aln->isStopCodon(state_map[i]))
+		state_map[i] = phylo_tree->aln->non_stop_codon[nt1*16+nt2*4+nt3];
+		if (phylo_tree->aln->isStopCodon(state_map[i]) || state_map[i] == STATE_INVALID)
 			outError("Stop codon encountered");
 		if (verbose_mode >= VB_MAX)
 			cout << " " << codons[i] << " " << state_map[i];
@@ -650,10 +661,10 @@ void ModelCodon::readCodonModel(istream &in, bool reset_params) {
 //			int id = col*(2*num_states-col-1)/2 + (row-col-1);
             double qentry = q[i*nscodons+j];
             int id = row*num_states+col;
-			assert(id < nrates && id >= 0);
+			ASSERT(id < nrates && id >= 0);
 			empirical_rates[id] = rates[id] = qentry;
             id = col*num_states+row;
-			assert(id < nrates && id >= 0);
+			ASSERT(id < nrates && id >= 0);
 			empirical_rates[id] = rates[id] = qentry;
 		}
 	}
@@ -681,9 +692,33 @@ void ModelCodon::readCodonModel(string &str, bool reset_params) {
 	}
 }
 
+void ModelCodon::readCodonModelFile(const char *filename, bool reset_params) {
+	try {
+		ifstream in;
+        // set the failbit and badbit
+        in.exceptions(ios::failbit | ios::badbit);
+        in.open(filename);
+        // remove the failbit
+        in.exceptions(ios::badbit);
+
+		readCodonModel(in, reset_params);
+
+        in.clear();
+        // set the failbit again
+        in.exceptions(ios::failbit | ios::badbit);
+        in.close();
+	}
+	catch (const char *str) {
+		outError(str);
+	}
+	catch (...) {
+        outError(ERR_READ_INPUT, filename);
+    }
+}
+
 void ModelCodon::decomposeRateMatrix() {
     computeCodonRateMatrix();
-    ModelGTR::decomposeRateMatrix();
+    ModelMarkov::decomposeRateMatrix();
 }
 
 void ModelCodon::computeCodonRateMatrix() {
@@ -874,7 +909,7 @@ bool ModelCodon::getVariables(double *variables) {
             changed |= (kappa2 != variables[j]);
             kappa2 = variables[j++];
         }
-        assert(j == num_params+1);
+        ASSERT(j == num_params+1);
     }
 	if (freq_type == FREQ_ESTIMATE) {
         // 2015-09-07: relax the sum of state_freq to be 1, this will be done at the end of optimization
@@ -913,7 +948,7 @@ void ModelCodon::setVariables(double *variables) {
         if (!fix_kappa2)
             variables[j++] = kappa2;
         
-		assert(j == num_params+1);
+		ASSERT(j == num_params+1);
 	}
 	if (freq_type == FREQ_ESTIMATE) {
         // 2015-09-07: relax the sum of state_freq to be 1, this will be done at the end of optimization
@@ -977,9 +1012,9 @@ double ModelCodon::optimizeParameters(double gradient_epsilon) {
 	// by BFGS algorithm
 	setVariables(variables);
 	setBounds(lower_bound, upper_bound, bound_check);
-//    if (phylo_tree->params->optimize_alg.find("BFGS-B") == string::npos)
-//        score = -minimizeMultiDimen(variables, ndim, lower_bound, upper_bound, bound_check, max(gradient_epsilon, TOL_RATE));
-//    else
+    if (phylo_tree->params->optimize_alg.find("BFGS-B") == string::npos)
+        score = -minimizeMultiDimen(variables, ndim, lower_bound, upper_bound, bound_check, max(gradient_epsilon, TOL_RATE));
+    else
         score = -L_BFGS_B(ndim, variables+1, lower_bound+1, upper_bound+1, max(gradient_epsilon, TOL_RATE));
 
 	bool changed = getVariables(variables);
